# Copyright 2021 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Tests for rdp_privacy_accountant."""

import math
import sys

from absl.testing import absltest
from absl.testing import parameterized
import mpmath
import numpy as np

from dp_accounting import dp_event
from dp_accounting import privacy_accountant
from dp_accounting import privacy_accountant_test
from dp_accounting.rdp import rdp_privacy_accountant


def _get_test_rdp(event, count=1):
  accountant = rdp_privacy_accountant.RdpAccountant(orders=[2.71828])
  accountant.compose(event, count)
  return accountant._rdp[0]


def _log_float_mp(x):
  # Convert multi-precision input to float log space.
  if x >= sys.float_info.min:
    return float(mpmath.log(x))
  else:
    return -np.inf


def _compute_a_mp(sigma, q, alpha):
  """Compute A_alpha for arbitrary alpha by numerical integration."""

  def mu0(x):
    return mpmath.npdf(x, mu=0, sigma=sigma)

  def _mu_over_mu0(x, q, sigma):
    return (1 - q) + q * mpmath.exp((2 * x - 1) / (2 * sigma**2))

  def a_alpha_fn(z):
    return mu0(z) * _mu_over_mu0(z, q, sigma) ** alpha

  bounds = (-mpmath.inf, mpmath.inf)
  a_alpha, _ = mpmath.quad(a_alpha_fn, bounds, error=True, maxdegree=8)
  return a_alpha


def _compose_trees(noise_multiplier, step_counts, orders):
  accountant = rdp_privacy_accountant.RdpAccountant(
      orders, privacy_accountant.NeighboringRelation.REPLACE_SPECIAL
  )
  accountant.compose(
      dp_event.ComposedDpEvent([
          dp_event.SingleEpochTreeAggregationDpEvent(
              noise_multiplier, step_count
          )
          for step_count in step_counts
      ])
  )
  return accountant


def _compose_trees_single_epoch(noise_multiplier, step_counts, orders):
  accountant = rdp_privacy_accountant.RdpAccountant(
      orders, privacy_accountant.NeighboringRelation.REPLACE_SPECIAL
  )
  accountant.compose(
      dp_event.SingleEpochTreeAggregationDpEvent(noise_multiplier, step_counts)
  )
  return accountant


class RdpPrivacyAccountantTest(
    privacy_accountant_test.PrivacyAccountantTest, parameterized.TestCase
):

  def _make_test_accountants(self):
    return [
        rdp_privacy_accountant.RdpAccountant(
            [2.0], privacy_accountant.NeighboringRelation.ADD_OR_REMOVE_ONE
        ),
        rdp_privacy_accountant.RdpAccountant(
            [2.0], privacy_accountant.NeighboringRelation.REPLACE_ONE
        ),
        rdp_privacy_accountant.RdpAccountant(
            [2.0], privacy_accountant.NeighboringRelation.REPLACE_SPECIAL
        ),
    ]

  def test_supports(self):
    aor_accountant = rdp_privacy_accountant.RdpAccountant(
        [2.0], privacy_accountant.NeighboringRelation.ADD_OR_REMOVE_ONE
    )
    ro_accountant = rdp_privacy_accountant.RdpAccountant(
        [2.0], privacy_accountant.NeighboringRelation.REPLACE_ONE
    )
    rs_accountant = rdp_privacy_accountant.RdpAccountant(
        [2.0], privacy_accountant.NeighboringRelation.REPLACE_SPECIAL
    )

    event = dp_event.GaussianDpEvent(1.0)
    self.assertTrue(aor_accountant.supports(event))
    self.assertTrue(ro_accountant.supports(event))

    event = dp_event.SelfComposedDpEvent(dp_event.GaussianDpEvent(1.0), 6)
    self.assertTrue(aor_accountant.supports(event))
    self.assertTrue(ro_accountant.supports(event))

    event = dp_event.ComposedDpEvent(
        [dp_event.GaussianDpEvent(1.0), dp_event.GaussianDpEvent(2.0)]
    )
    self.assertTrue(aor_accountant.supports(event))
    self.assertTrue(ro_accountant.supports(event))

    event = dp_event.PoissonSampledDpEvent(0.1, dp_event.GaussianDpEvent(1.0))
    self.assertTrue(aor_accountant.supports(event))
    self.assertFalse(ro_accountant.supports(event))

    composed_gaussian = dp_event.ComposedDpEvent(
        [dp_event.GaussianDpEvent(1.0), dp_event.GaussianDpEvent(2.0)]
    )
    event = dp_event.PoissonSampledDpEvent(0.1, composed_gaussian)
    self.assertTrue(aor_accountant.supports(event))
    self.assertFalse(ro_accountant.supports(event))

    event = dp_event.SampledWithoutReplacementDpEvent(
        1000, 10, dp_event.GaussianDpEvent(1.0)
    )
    self.assertFalse(aor_accountant.supports(event))
    self.assertTrue(ro_accountant.supports(event))

    event = dp_event.SampledWithoutReplacementDpEvent(
        1000, 10, composed_gaussian
    )
    self.assertFalse(aor_accountant.supports(event))
    self.assertTrue(ro_accountant.supports(event))

    event = dp_event.SampledWithReplacementDpEvent(
        1000, 10, dp_event.GaussianDpEvent(1.0)
    )
    self.assertFalse(aor_accountant.supports(event))
    self.assertFalse(ro_accountant.supports(event))

    event = dp_event.RandomizedResponseDpEvent(0.3, 15)
    self.assertTrue(rs_accountant.supports(event))
    self.assertTrue(ro_accountant.supports(event))
    self.assertFalse(aor_accountant.supports(event))

  def test_rdp_composition(self):
    base_event = dp_event.GaussianDpEvent(3.14159)
    base_rdp = _get_test_rdp(base_event)

    rdp_with_count = _get_test_rdp(base_event, count=6)
    self.assertAlmostEqual(rdp_with_count, base_rdp * 6)

    rdp_with_self_compose = _get_test_rdp(
        dp_event.SelfComposedDpEvent(base_event, 6)
    )
    self.assertAlmostEqual(rdp_with_self_compose, base_rdp * 6)

    rdp_with_self_compose_and_count = _get_test_rdp(
        dp_event.SelfComposedDpEvent(base_event, 2), count=3
    )
    self.assertAlmostEqual(rdp_with_self_compose_and_count, base_rdp * 6)

    rdp_with_compose = _get_test_rdp(dp_event.ComposedDpEvent([base_event] * 6))
    self.assertAlmostEqual(rdp_with_compose, base_rdp * 6)

    rdp_with_compose_and_self_compose = _get_test_rdp(
        dp_event.ComposedDpEvent([
            dp_event.SelfComposedDpEvent(base_event, 1),
            dp_event.SelfComposedDpEvent(base_event, 2),
            dp_event.SelfComposedDpEvent(base_event, 3),
        ])
    )
    self.assertAlmostEqual(rdp_with_compose_and_self_compose, base_rdp * 6)

    base_event_2 = dp_event.GaussianDpEvent(1.61803)
    base_rdp_2 = _get_test_rdp(base_event_2)
    rdp_with_heterogeneous_compose = _get_test_rdp(
        dp_event.ComposedDpEvent([base_event, base_event_2])
    )
    self.assertAlmostEqual(
        rdp_with_heterogeneous_compose, base_rdp + base_rdp_2
    )

  @parameterized.parameters(
      dp_event.PoissonSampledDpEvent(0, dp_event.GaussianDpEvent(1.0)),
      dp_event.PoissonSampledDpEvent(0, dp_event.GaussianDpEvent(0.0)),
  )
  def test_zero_poisson_sample(self, event):
    accountant = rdp_privacy_accountant.RdpAccountant([3.14159])
    accountant.compose(event)
    self.assertEqual(accountant.get_epsilon(1e-10), 0)
    self.assertEqual(accountant.get_delta(1e-10), 0)

  def test_zero_fixed_batch_sample(self):
    accountant = rdp_privacy_accountant.RdpAccountant(
        [3.14159], privacy_accountant.NeighboringRelation.REPLACE_ONE
    )
    accountant.compose(
        dp_event.SampledWithoutReplacementDpEvent(
            1000, 0, dp_event.GaussianDpEvent(1.0)
        )
    )
    self.assertEqual(accountant.get_epsilon(1e-10), 0)
    self.assertEqual(accountant.get_delta(1e-10), 0)

  def test_epsilon_non_private_gaussian(self):
    accountant = rdp_privacy_accountant.RdpAccountant([3.14159])
    accountant.compose(dp_event.GaussianDpEvent(0))
    self.assertEqual(accountant.get_epsilon(1e-1), np.inf)

  def test_compute_rdp_gaussian(self):
    alpha = 3.14159
    sigma = 2.71828
    event = dp_event.GaussianDpEvent(sigma)
    accountant = rdp_privacy_accountant.RdpAccountant(orders=[alpha])
    accountant.compose(event)
    self.assertAlmostEqual(accountant._rdp[0], alpha / (2 * sigma**2))

  def test_compute_rdp_multi_gaussian(self):
    alpha = 3.14159
    sigma1, sigma2 = 2.71828, 6.28319

    rdp1 = alpha / (2 * sigma1**2)
    rdp2 = alpha / (2 * sigma2**2)
    rdp = rdp1 + rdp2

    accountant = rdp_privacy_accountant.RdpAccountant(orders=[alpha])
    accountant.compose(
        dp_event.PoissonSampledDpEvent(
            1.0,
            dp_event.ComposedDpEvent([
                dp_event.GaussianDpEvent(sigma1),
                dp_event.GaussianDpEvent(sigma2),
            ]),
        )
    )
    self.assertAlmostEqual(accountant._rdp[0], rdp)

  def test_effective_gaussian_noise_multiplier_basic(self):
    sigma = 2.71828
    event = dp_event.GaussianDpEvent(sigma)
    sigma_out = rdp_privacy_accountant._effective_gaussian_noise_multiplier(
        event
    )
    self.assertEqual(sigma_out, sigma)

  def test_effective_gaussian_noise_multiplier_composed(self):
    np.random.seed(0xBAD5EED)
    sigmas = np.random.uniform(size=(4,))

    event = dp_event.ComposedDpEvent([
        dp_event.GaussianDpEvent(sigmas[0]),
        dp_event.SelfComposedDpEvent(dp_event.GaussianDpEvent(sigmas[1]), 3),
        dp_event.ComposedDpEvent([
            dp_event.GaussianDpEvent(sigmas[2]),
            dp_event.GaussianDpEvent(sigmas[3]),
        ]),
    ])

    sigma = rdp_privacy_accountant._effective_gaussian_noise_multiplier(event)
    multi_sigmas = list(sigmas) + [sigmas[1]] * 2
    expected = sum(s**-2 for s in multi_sigmas) ** -0.5
    self.assertAlmostEqual(sigma, expected)

  _LAPLACE_EVENT = dp_event.LaplaceDpEvent(1.0)

  @parameterized.named_parameters(
      ('simple', _LAPLACE_EVENT),
      ('composed', dp_event.ComposedDpEvent([_LAPLACE_EVENT])),
      ('self_composed', dp_event.SelfComposedDpEvent(_LAPLACE_EVENT, 3)),
  )
  def test_effective_gaussian_noise_multiplier_invalid_event(self, event):
    result = rdp_privacy_accountant._effective_gaussian_noise_multiplier(event)
    self.assertEqual(result, self._LAPLACE_EVENT)

  def test_compute_rdp_poisson_sampled_gaussian(self):
    orders = [1.5, 2.5, 5, 50, 100, np.inf]
    noise_multiplier = 2.5
    sampling_probability = 0.01
    count = 50
    event = dp_event.SelfComposedDpEvent(
        dp_event.PoissonSampledDpEvent(
            sampling_probability, dp_event.GaussianDpEvent(noise_multiplier)
        ),
        count,
    )
    accountant = rdp_privacy_accountant.RdpAccountant(orders=orders)
    accountant.compose(event)
    self.assertTrue(
        np.allclose(
            accountant._rdp,
            [
                6.70579741e-04,
                1.08548710e-03,
                2.18075656e-03,
                2.38460375e-02,
                1.67416308e02,
                np.inf,
            ],
            rtol=1e-4,
        )
    )

  @parameterized.named_parameters(
      (
          'small_eps',
          0.3,
          10,
          [
              1.5030552584146897,
              1.7801182478869704,
              1.9091969256833639,
              1.9814516798923105,
              1.9846954517418232,
              1.9878743481543455,
          ],
      ),
      (
          'medium_eps',
          0.05,
          20,
          [
              2.8985312222056674,
              2.9615414620514047,
              2.982908949957726,
              2.994685471627712,
              2.9952141594692665,
              2.995732273553991,
          ],
      ),
      (
          'large_eps',
          0.01,
          3,
          [
              3.851345343526459,
              4.334924907685137,
              4.503803908976894,
              4.596895387863428,
              4.601074578835478,
              4.605170185988092,
          ],
      ),
  )
  def test_compute_rdp_randomized_response_replace_special(
      self, noise_parameter, num_buckets, expected_rdps
  ):
    orders = [1.5, 2.5, 5, 50, 100, np.inf]
    event = dp_event.RandomizedResponseDpEvent(noise_parameter, num_buckets)
    accountant = rdp_privacy_accountant.RdpAccountant(
        orders,
        privacy_accountant.NeighboringRelation.REPLACE_SPECIAL,
    )
    accountant.compose(event)
    self.assertTrue(
        np.allclose(
            accountant._rdp,
            expected_rdps,
            rtol=1e-4,
        )
    )

  @parameterized.named_parameters(
      (
          'small_eps',
          0.3,
          10,
          [
              2.6946320226546314,
              2.983865364028855,
              3.11316970070507,
              3.1854244842182466,
              3.1886682560677593,
              3.1918471524802814,
          ],
      ),
      (
          'medium_eps',
          0.05,
          20,
          [
              5.850317703374384,
              5.910360162307462,
              5.930633082131939,
              5.94180620835157,
              5.942307807732949,
              5.942799375126701,
          ],
      ),
      (
          'large_eps',
          0.01,
          3,
          [
              5.684126770932727,
              5.692634596034571,
              5.695421239467805,
              5.696956976543137,
              5.697025920968522,
              5.697093486505405,
          ],
      ),
  )
  def test_compute_rdp_randomized_response_replace_one(
      self, noise_parameter, num_buckets, expected_rdps
  ):
    orders = [1.5, 2.5, 5, 50, 100, np.inf]
    event = dp_event.RandomizedResponseDpEvent(noise_parameter, num_buckets)
    accountant = rdp_privacy_accountant.RdpAccountant(
        orders,
        privacy_accountant.NeighboringRelation.REPLACE_ONE,
    )
    accountant.compose(event)
    self.assertTrue(
        np.allclose(
            accountant._rdp,
            expected_rdps,
            rtol=1e-4,
        )
    )

  @parameterized.named_parameters(
      ('noise_param_1', 1, 2, [1.1, 1.2], [0, 0]),
      ('noise_param_0', 0.0, 2, [1.1, 1.2], [np.inf, np.inf]),
      ('num_buckets_1', 0.1, 1, [1.1, 1.2], [0, 0]),
  )
  def test_compute_rdp_randomized_response_boundary(
      self, noise_param, num_buckets, alphas, expected_rdps
  ):
    for rel in [
        privacy_accountant.NeighboringRelation.REPLACE_SPECIAL,
        privacy_accountant.NeighboringRelation.REPLACE_ONE,
    ]:
      self.assertSequenceAlmostEqual(
          rdp_privacy_accountant._compute_randomized_response_rdp(
              noise_param, num_buckets, alphas, rel
          ),
          expected_rdps,
      )

  @parameterized.named_parameters(
      (
          'replace_special',
          privacy_accountant.NeighboringRelation.REPLACE_SPECIAL,
          2.302585092994046,
      ),
      (
          'replace_one',
          privacy_accountant.NeighboringRelation.REPLACE_ONE,
          4.51085950651685,
      ),
  )
  def test_compute_rdp_randomized_response_inf_order(self, rel, expected_rdp):
    noise_param = 0.1
    num_buckets = 10
    orders = [np.inf]
    self.assertSequenceAlmostEqual(
        rdp_privacy_accountant._compute_randomized_response_rdp(
            noise_param, num_buckets, orders, rel
        ),
        [expected_rdp],
    )

  def test_compute_rdp_randomized_response_raises(self):
    with self.assertRaisesRegex(ValueError, 'noise_parameter must be in'):
      for rel in [
          privacy_accountant.NeighboringRelation.REPLACE_SPECIAL,
          privacy_accountant.NeighboringRelation.REPLACE_ONE,
      ]:
        rdp_privacy_accountant._compute_randomized_response_rdp(
            -1, 10, [1.1, 1.2], rel
        )
      with self.assertRaisesRegex(ValueError, 'noise_parameter must be in'):
        rdp_privacy_accountant._compute_randomized_response_rdp(
            2, 10, [1.1, 1.2], rel
        )
      with self.assertRaisesRegex(ValueError, 'num_buckets must be >= 1'):
        rdp_privacy_accountant._compute_randomized_response_rdp(
            0.1, 0, [1.1, 1.2], rel
        )
      with self.assertRaisesRegex(
          ValueError, 'Renyi divergence order alpha must be > 1'
      ):
        rdp_privacy_accountant._compute_randomized_response_rdp(
            0.1, 10, [1, 1.2], rel
        )

  def test_compute_epsilon_delta_pure_dp(self):
    orders = range(2, 33)
    rdp = [1.1 for _ in orders]  # Constant corresponds to pure DP.

    epsilon, optimal_order = rdp_privacy_accountant.compute_epsilon(
        orders, rdp, delta=1e-5
    )
    # Compare with epsilon computed by hand.
    self.assertAlmostEqual(epsilon, 1.32783806176)
    self.assertEqual(optimal_order, 32)

    delta, optimal_order = rdp_privacy_accountant.compute_delta(
        orders, rdp, epsilon=1.32783806176
    )
    self.assertAlmostEqual(delta, 1e-5)
    self.assertEqual(optimal_order, 32)

  def test_compute_epsilon_delta_gaussian(self):
    orders = [0.001 * i for i in range(1000, 100000)]

    # noise multiplier is chosen to obtain exactly (1,1e-6)-DP.
    rdp = rdp_privacy_accountant._compute_rdp_poisson_subsampled_gaussian(
        1, 4.530877117, orders
    )

    eps = rdp_privacy_accountant.compute_epsilon(orders, rdp, delta=1e-6)[0]
    self.assertAlmostEqual(eps, 1)

    delta = rdp_privacy_accountant.compute_delta(orders, rdp, epsilon=1)[0]
    self.assertAlmostEqual(delta, 1e-6)

  params = (
      {'q': 1e-7, 'sigma': 0.1, 'order': 1.01},
      {'q': 1e-6, 'sigma': 0.1, 'order': 256},
      {'q': 1e-5, 'sigma': 0.1, 'order': 256.1},
      {'q': 1e-6, 'sigma': 1, 'order': 27},
      {'q': 1e-4, 'sigma': 1.0, 'order': 1.5},
      {'q': 1e-3, 'sigma': 1.0, 'order': 2},
      {'q': 0.01, 'sigma': 10, 'order': 20},
      {'q': 0.1, 'sigma': 100, 'order': 20.5},
      {'q': 0.99, 'sigma': 0.1, 'order': 256},
      {'q': 0.999, 'sigma': 100, 'order': 256.1},
  )

  # pylint:disable=undefined-variable
  @parameterized.parameters(p for p in params)
  def test_compute_log_a_equals_mp(self, q, sigma, order):
    # Compare the cheap computation of log(A) with an expensive, multi-precision
    # computation.
    log_a = rdp_privacy_accountant._compute_log_a(q, sigma, order)
    log_a_mp = _log_float_mp(_compute_a_mp(sigma, q, order))
    np.testing.assert_allclose(log_a, log_a_mp, rtol=1e-4, atol=1e-8)

  def test_delta_bounds_gaussian(self):
    # Compare the optimal bound for Gaussian with the one derived from RDP.
    # Also compare the RDP upper bound with the "standard" upper bound.
    orders = [0.1 * x for x in range(10, 505)]
    eps_vec = [0.1 * x for x in range(500)]
    rdp = rdp_privacy_accountant._compute_rdp_poisson_subsampled_gaussian(
        1, 1, orders
    )
    for eps in eps_vec:
      delta = rdp_privacy_accountant.compute_delta(orders, rdp, epsilon=eps)[0]
      # For comparison, we compute the optimal guarantee for Gaussian
      # using https://arxiv.org/abs/1805.06530 Theorem 8 (in v2).
      delta0 = math.erfc((eps - 0.5) / math.sqrt(2)) / 2
      delta0 = (
          delta0 - math.exp(eps) * math.erfc((eps + 0.5) / math.sqrt(2)) / 2
      )
      self.assertLessEqual(delta0, delta + 1e-300)  # need tolerance 10^-300

      # Compute the "standard" upper bound, which should be an upper bound.
      # Note, if orders is too sparse, this will NOT be an upper bound.
      if eps >= 0.5:
        delta1 = math.exp(-0.5 * (eps - 0.5) ** 2)
      else:
        delta1 = 1
      self.assertLessEqual(delta, delta1 + 1e-300)

  def test_epsilon_delta_consistency(self):
    orders = range(2, 50)  # Large range of orders (helps test for overflows).
    for q in [0, 0.01, 0.1, 0.8, 1.0]:
      for multiplier in [0.0, 0.1, 1.0, 10.0, 100.0]:
        event = dp_event.PoissonSampledDpEvent(
            q, dp_event.GaussianDpEvent(multiplier)
        )
        accountant = rdp_privacy_accountant.RdpAccountant(orders)
        accountant.compose(event)
        for delta in [0.99, 0.9, 0.1, 0.01, 1e-3, 1e-5, 1e-9, 1e-12]:
          epsilon = accountant.get_epsilon(delta)
          delta2 = accountant.get_delta(epsilon)
          if np.isposinf(epsilon):
            self.assertEqual(delta2, 1.0)
          elif epsilon == 0:
            self.assertLessEqual(delta2, delta)
          else:
            self.assertAlmostEqual(delta, delta2)

  @parameterized.named_parameters(
      ('add_remove', privacy_accountant.NeighboringRelation.ADD_OR_REMOVE_ONE),
      ('replace', privacy_accountant.NeighboringRelation.REPLACE_ONE),
  )
  def test_tree_wrong_neighbor_rel(self, neighboring_relation):
    event = dp_event.SingleEpochTreeAggregationDpEvent(1.0, 1)
    accountant = rdp_privacy_accountant.RdpAccountant(
        neighboring_relation=neighboring_relation
    )
    self.assertFalse(accountant.supports(event))

  @parameterized.named_parameters(('eps20', 1.13, 19.74), ('eps2', 8.83, 2.04))
  def test_compute_eps_tree(self, noise_multiplier, eps):
    orders = [1 + x / 10 for x in range(1, 100)] + list(range(12, 64))
    # This test is based on the StackOverflow setting in "Practical and
    # Private (Deep) Learning without Sampling or Shuffling". The calculated
    # epsilon could be better as the method in this package keeps improving.
    step_counts, target_delta = 1600, 1e-6
    new_eps = _compose_trees_single_epoch(
        noise_multiplier, step_counts, orders
    ).get_epsilon(target_delta)
    self.assertLess(new_eps, eps)

  @parameterized.named_parameters(
      ('restart4', [400] * 4),
      ('restart2', [800] * 2),
      ('adaptive', [10, 400, 400, 400, 390]),
  )
  def test_compose_tree_rdp(self, step_counts):
    noise_multiplier, orders = 0.1, [1]

    def get_rdp(step_count):
      return _compose_trees_single_epoch(
          noise_multiplier, [step_count], orders
      )._rdp[0]

    rdp_summed = sum(get_rdp(step_count) for step_count in step_counts)
    rdp_composed = _compose_trees(noise_multiplier, step_counts, orders)._rdp[0]
    self.assertTrue(np.allclose(rdp_composed, rdp_summed, rtol=1e-12))

  def test_single_epoch_multi_tree_rdp(self):
    noise_multiplier, orders = 0.1, [1]
    step_counts = [10, 40, 30, 20]
    single_rdp = _compose_trees_single_epoch(
        noise_multiplier, step_counts, orders
    )._rdp[0]

    max_rdp = max(
        _compose_trees_single_epoch(noise_multiplier, step_count, orders)._rdp[
            0
        ]
        for step_count in step_counts
    )

    self.assertEqual(single_rdp, max_rdp)

  @parameterized.named_parameters(
      ('restart4', [400] * 4),
      ('restart2', [800] * 2),
      ('adaptive', [10, 400, 400, 400, 390]),
  )
  def test_compute_eps_tree_decreasing(self, step_counts):
    # Test privacy epsilon decreases with noise multiplier increasing when
    # keeping other parameters the same.
    orders = [1 + x / 10.0 for x in range(1, 100)] + list(range(12, 64))
    target_delta = 1e-6
    prev_eps = np.inf
    for noise_multiplier in [0.1 * x for x in range(1, 100, 5)]:
      accountant = _compose_trees(noise_multiplier, step_counts, orders)
      eps = accountant.get_epsilon(target_delta=target_delta)
      self.assertLess(eps, prev_eps)
      prev_eps = eps

  @parameterized.named_parameters(
      ('negative_noise', -1, [3]),
      ('negative_steps', 1, [-3]),
  )
  def test_compute_rdp_tree_restart_raise(self, noise_multiplier, step_counts):
    with self.assertRaisesRegex(ValueError, 'non-negative'):
      _compose_trees(noise_multiplier, step_counts, orders=[1])

  @parameterized.named_parameters(
      ('t100n0.1', 100, 0.1),
      ('t1000n0.01', 1000, 0.01),
  )
  def test_no_tree_no_sampling(self, total_steps, noise_multiplier):
    orders = [1 + x / 10 for x in range(1, 100)] + list(range(12, 64))
    tree_rdp = _compose_trees(noise_multiplier, [1] * total_steps, orders)._rdp
    accountant = rdp_privacy_accountant.RdpAccountant(orders)
    event = dp_event.SelfComposedDpEvent(
        dp_event.GaussianDpEvent(noise_multiplier), total_steps
    )
    accountant.compose(event)
    base_rdp = accountant._rdp
    self.assertTrue(np.allclose(tree_rdp, base_rdp, rtol=1e-12))

  @parameterized.named_parameters(
      ('small_eps', 0.01, 1),
      ('medium_eps', 1.0, 1),
      ('large_eps', 100.0, 1),
      ('repetition', 1.0, 100),
  )
  def test_laplace(self, eps, count):
    event = dp_event.LaplaceDpEvent(1 / eps)
    if count != 1:
      event = dp_event.SelfComposedDpEvent(event, count)
    # Simulate Pure DP by using a large Renyi order.
    accountant = rdp_privacy_accountant.RdpAccountant(
        orders=[1.0, 1 + 1e-6, 1.1 - 1e-6, 1.1 + 1e-6, 1e10]
    )
    accountant.compose(event)
    # Check basic composition by having small delta.
    self.assertAlmostEqual(accountant.get_epsilon(1e-10), eps * count)
    # Check KL divergence, a.k.a. expected privacy loss, a.k.a. order=1.
    self.assertLessEqual(accountant._rdp[0], min(eps, eps * eps / 2) * count)
    # Check consistency between the three formulas.
    self.assertAlmostEqual(accountant._rdp[0], accountant._rdp[1], places=3)
    self.assertAlmostEqual(accountant._rdp[2], accountant._rdp[3], places=3)

  # The function _truncated_negative_binomial_mean computes the mean in
  # multiple ways to ensure numerical stability.
  # This test checks that those different ways of computing are consistent.
  @parameterized.named_parameters(
      ('gamma_shape0', 0.9, 0, 0.9 - 1e-9, 0),
      ('gamma_shape2', 0.9, 2, 0.9 - 1e-9, 2),
      ('gamma_shape_0.5', 0.9, 0.5, 0.9 - 1e-9, 0.5),
      (
          'x_shape2',
          math.exp(-0.05),
          2,
          math.exp(-0.05) - 1e-9,
          2,
      ),  # x = shape * math.log(gamma) = -0.1
      (
          'x_shape0.5',
          math.exp(-0.2),
          0.5,
          math.exp(-0.2) - 1e-9,
          0.5,
      ),  # x = shape * math.log(gamma) = -0.1
      ('shape_0', 0.6, 0, 0.6, 1e-9),
      ('shape_1', 0.6, 1, 0.6, 1 + 1e-9),
  )
  def test_truncated_negative_binomial_mean(
      self, gamma1, shape1, gamma2, shape2
  ):
    mean1 = rdp_privacy_accountant._truncated_negative_binomial_mean(
        gamma1, shape1
    )
    mean2 = rdp_privacy_accountant._truncated_negative_binomial_mean(
        gamma2, shape2
    )
    self.assertAlmostEqual(mean1, mean2)

  @parameterized.named_parameters(
      ('1e-7', 1e-7), ('.1', 0.1), ('0.999999', 1 - 1e-6), ('1', 1)
  )
  def test_truncated_negative_binomial_mean2(self, gamma):
    # Test this function by simply applying the non-numerically stable formula.
    # logarithmic distribution
    mean = rdp_privacy_accountant._truncated_negative_binomial_mean(gamma, 0)
    if gamma == 1:
      ans = 1
    else:
      ans = (1 - 1 / gamma) / math.log(gamma)
    self.assertAlmostEqual(mean, ans)

    # geometric Distribution
    mean = rdp_privacy_accountant._truncated_negative_binomial_mean(gamma, 1)
    self.assertAlmostEqual(mean, 1 / gamma)

    # general TNB Distribution
    for shape in [0.01, 0.5, 0.99, 1.01, 2, 10]:
      mean = rdp_privacy_accountant._truncated_negative_binomial_mean(
          gamma, shape
      )
      if gamma == 1:
        ans = 1
      else:
        ans = shape * (1 / gamma - 1) / (1 - gamma**shape)
      self.assertAlmostEqual(mean, ans)

  # _gamma_truncated_negative_binomial is meant to be the inverse of
  # _truncated_negative_binomial_mean, so we test this.
  @parameterized.named_parameters(
      ('shape0a', 0.1, 0),
      ('shape0.5a', 0.1, 0.5),
      ('shape1a', 0.1, 1),
      ('shape2a', 0.1, 2),
      ('shape0b', 0.0001, 0),
      ('shape0.5b', 0.0001, 0.5),
      ('shape1b', 0.0001, 1),
      ('shape2b', 0.0001, 2),
      ('shape0c', 1, 0),
      ('shape0.5c', 1, 0.5),
      ('shape1c', 1, 1),
      ('shape2c', 1, 2),
      ('shape0', 0.999, 0),
      ('shape0.5', 0.999, 0.5),
      ('shape1', 0.999, 1),
      ('shape2', 0.999, 2),
  )
  def test_gamma_truncated_negative_binomial(self, gamma, shape):
    mean = rdp_privacy_accountant._truncated_negative_binomial_mean(
        gamma, shape
    )
    g = rdp_privacy_accountant._gamma_truncated_negative_binomial(shape, mean)
    self.assertAlmostEqual(g, gamma)

  @parameterized.named_parameters(
      ('logarithmic', 1, 1000, 0),
      ('geometric', 1, 1000, 1),
      ('negative binomial 0.5', 1, 1000, 0.5),
      ('negative binomial 2', 1, 100, 2),
      ('negative binomial 5', 1, 1000, 5),
  )
  def test_repeat_select_pure_negative_binomial(self, eps, mean, shape):
    # Test the Repeat and Select DP event in the almost-pure DP case.
    event = dp_event.LaplaceDpEvent(1 / eps)
    event = dp_event.RepeatAndSelectDpEvent(event, mean, shape)
    # Use single large order to simulate pure DP.
    accountant = rdp_privacy_accountant.RdpAccountant(orders=[1e10])
    accountant.compose(event)
    # Correct answer is given by Corollary 3 https://arxiv.org/abs/2110.03620
    self.assertAlmostEqual(accountant._rdp[0], eps * (2 + shape))
    self.assertAlmostEqual(accountant.get_epsilon(1e-10), eps * (2 + shape))

  @parameterized.named_parameters(
      ('shape0', 0, 1),
      ('shape0.5', 0.5, 10),
      ('shape1', 1, 0.1),
      ('shape2', 2, 1),
  )
  def test_repeat_select_trivial(self, shape, sigma):
    # Test the repeat and select function in the trivial mean=1 case.
    orders = [
        1,
        1 + 1e-6,  # We include 1, as otherwise this test fails.
        2,
        3,
        4,
        5,
        6,
        7,
        8,
        9,
        10,
        12,
        14,
        16,
        20,
        24,
        28,
        32,
        48,
        64,
        128,
        256,
        512,
        1024,
    ]
    event1 = dp_event.GaussianDpEvent(sigma)
    accountant1 = rdp_privacy_accountant.RdpAccountant(orders=orders)
    accountant1.compose(event1)
    event2 = dp_event.RepeatAndSelectDpEvent(event1, 1, shape)
    accountant2 = rdp_privacy_accountant.RdpAccountant(orders=orders)
    accountant2.compose(event2)
    for i in range(len(accountant1._orders)):
      if orders[i] > 1:  # Otherwise our formula doesn't work.
        self.assertAlmostEqual(accountant1._rdp[i], accountant2._rdp[i])

  @parameterized.named_parameters(
      ('small0', 0.01, 0.01, 0),
      ('med0', 1, 0.1, 0),
      ('large0', 10, 0.99, 0),
      ('small0.5', 0.01, 0.01, 0.5),
      ('med0.5', 1, 0.1, 0.5),
      ('large0.5', 10, 0.99, 0.5),
      ('small1', 0.01, 0.01, 1),
      ('med1', 1, 0.1, 1),
      ('large1', 10, 0.99, 1),
      ('small5', 0.01, 0.01, 5),
      ('med5', 1, 0.1, 5),
      ('large5', 10, 0.99, 5),
  )
  def test_repeat_select_gaussian_negative_binomial(self, rho, gamma, shape):
    # Test the Repeat and Select DP event in the Gaussian case.
    # Correct answer is given by Corollary 4 https://arxiv.org/abs/2110.03620
    mean = rdp_privacy_accountant._truncated_negative_binomial_mean(
        gamma, shape
    )
    rho = min(rho, -math.log(gamma))  # We need rho<=log(1/gamma).
    self.assertGreater(rho, 0)  # Otherwise we get division by zero.
    orders = [
        1,
        1.1,
        2,
        math.sqrt(-math.log(gamma) / rho),
        1 + math.sqrt(math.log(mean) / rho),
        3,
        5,
        10,
        100,
        1000,
        10000,
    ]
    event = dp_event.GaussianDpEvent(math.sqrt(0.5 / rho))
    event = dp_event.RepeatAndSelectDpEvent(event, mean, shape)
    accountant = rdp_privacy_accountant.RdpAccountant(orders=orders)
    accountant.compose(event)
    for i in range(len(orders)):
      order = accountant._orders[i]
      rdp = accountant._rdp[i]
      if order <= 1 + math.sqrt(math.log(mean) / rho):
        eps = (
            2 * math.sqrt(rho * math.log(mean))
            + 2 * (1 + shape) * math.sqrt(-rho * math.log(gamma))
            - shape * rho
        )
      else:
        eps = (
            rho * (order - 1)
            + math.log(mean) / (order - 1)
            + 2 * (1 + shape) * math.sqrt(-rho * math.log(gamma))
            - shape * rho
        )
      self.assertAlmostEqual(rdp, eps, msg='order=' + str(order))

  @parameterized.named_parameters(
      ('mean1', 1, 1),
      ('mean2', 0.1, 2),
      ('mean10', 10, 10),
      ('mean100', 0.001, 100),
      ('mean10^4', 2, 1000),
      ('mean10^10', 1, 1e10),
  )
  def test_repeat_and_select_pure_poisson(self, eps, mean):
    event = dp_event.LaplaceDpEvent(1 / eps)
    event = dp_event.RepeatAndSelectDpEvent(event, mean, np.inf)
    alpha = 1 + 1 / math.expm1(eps)
    orders = [alpha, 1e10, 1e100]
    accountant = rdp_privacy_accountant.RdpAccountant(orders=orders)
    accountant.compose(event)
    ans = (
        eps
        + math.log1p(
            math.expm1((1 - 2 * alpha) * eps) * (alpha - 1) / (2 * alpha - 1)
        )
        / (alpha - 1)
        + math.log(mean) * math.expm1(eps)
    )
    self.assertAlmostEqual(accountant._orders[0], alpha)
    self.assertAlmostEqual(accountant._rdp[0] / ans, 1, places=3)

  @parameterized.named_parameters(('q_neg', -0.1), ('q_large', 1.1))
  def test_raises_on_invalid_sampling_rate(self, q):
    with self.assertRaisesRegex(ValueError, 'Sampling rate'):
      rdp_privacy_accountant.RdpAccountant().compose(
          dp_event.PoissonSampledDpEvent(q, dp_event.GaussianDpEvent(1))
      )

  def test_raises_on_invalid_noise_multiplier(self):
    with self.assertRaisesRegex(ValueError, 'Noise multiplier'):
      rdp_privacy_accountant.RdpAccountant().compose(
          dp_event.GaussianDpEvent(-1)
      )

  @parameterized.named_parameters(
      ('small_small', 0.001, 1),
      ('small_med', 0.001, 1000),
      ('small_large', 0.001, 1e9),
      ('med_small', 1, 1),
      ('med_med', 1, 1000),
      ('med_large', 1, 1e9),
      ('large_small', 1000, 1),
      ('large_med', 1000, 1000),
      ('large_large', 1000, 1e9),
  )
  def test_repeat_and_select_gaussian_poisson(self, sigma, mean):
    event = dp_event.GaussianDpEvent(sigma)
    event = dp_event.RepeatAndSelectDpEvent(event, mean, np.inf)
    accountant = rdp_privacy_accountant.RdpAccountant()
    accountant.compose(event)
    orders = accountant._orders
    rdp = []
    for order in orders:
      if order <= 1:  # Avoid division by zero.
        rdp.append(np.inf)
        continue
      eps = math.log1p(1 / (order - 1))
      x = (eps * sigma - 0.5 / sigma) / math.sqrt(2)
      y = (eps * sigma + 0.5 / sigma) / math.sqrt(2)
      delta = math.erfc(x) / 2 - math.exp(eps) * math.erfc(y) / 2
      rdp.append(
          order * 0.5 / (sigma**2) + mean * delta + math.log(mean) / (order - 1)
      )
    for order, accountant_rdp in zip(orders, accountant._rdp):
      lb = min(rdp[j] for j in range(len(orders)) if orders[j] >= order)
      self.assertLessEqual(lb, accountant_rdp)

  def test_log_a_frac_positive(self):
    # Testing a combination of q, sigma and alpha that formerly returned a
    # negative log_a_frac.
    for order in np.linspace(58.5, 59.5, 21):
      log_a = rdp_privacy_accountant._compute_log_a_frac(0.4, 12, order)
      self.assertGreater(log_a, 0)

  def test_log_a_frac_early_termination(self):
    # Test an event that is known to not converge for small orders.
    event = dp_event.PoissonSampledDpEvent(0.1, dp_event.GaussianDpEvent(1.0))
    accountant = rdp_privacy_accountant.RdpAccountant()
    with self.assertLogs(level='WARNING') as log:
      accountant.compose(event)
    self.assertNotEmpty([l for l in log.output if 'failed to converge' in l])
    self.assertIn(np.inf, accountant._rdp)


if __name__ == '__main__':
  absltest.main()



# Copyright 2022 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for test_util."""

import unittest

from absl.testing import parameterized

from dp_accounting.pld import test_util


class TestUtilTest(parameterized.TestCase):

  @parameterized.parameters(
      # Dictionary contained
      ({
          1: 0.1,
          2: 0.3
      }, {
          1: 0.1,
          2: 0.3,
          4: 0.2
      }, True),
      ({
          1: 1e-10,
          2: 0.3
      }, {
          2: 0.3,
          4: 0.2
      }, True),
      ({
          1.0: 0.1 + 1e-10,
          2.0: 0.3
      }, {
          1.0 + 1e-10: 0.1,
          2.0: 0.3 + 1e-10,
          4.0: 0.2
      }, True),
      # Dictionary not contained
      ({
          1: 0.1,
          2: 0.3
      }, {
          2: 0.3,
          4: 0.2
      }, False))
  def test_assert_dictionary_contained(self, dict1, dict2, expected_result):
    if expected_result:
      test_util.assert_dictionary_contained(self, dict1, dict2)
    else:
      with self.assertRaises(AssertionError):
        test_util.assert_dictionary_contained(self, dict1, dict2)

  @parameterized.parameters(
      # Key missing
      ({
          1: 0.1,
          2: 0.3
      }, {
          1: 0.1,
          3: 0.3,
      }, True),
      # Value not matching
      ({
          1: 0.1,
          2: 0.3
      }, {
          1: 0.1,
          2: 0.2
      }, False))
  def test_assert_dictionary_contained_error_messages(
      self, dict1, dict2, key_missing):
    with self.assertRaises(AssertionError) as cm:
      test_util.assert_dictionary_contained(self, dict1, dict2)
    if key_missing:
      self.assertStartsWith(str(cm.exception), 'False is not true : Key')
    else:
      self.assertStartsWith(str(cm.exception), 'False is not true : Value')

  @parameterized.parameters(
      # Dictionary almost equal
      ({
          1: 0.1,
          2: 0.3,
      }, {
          1: 0.1,
          2: 0.3
      }, True),
      ({
          1: 1e-10,
          2: 0.3,
          4: 0.2,
      }, {
          2: 0.3,
          4: 0.2
      }, True),
      ({
          1.0: 0.1 + 1e-10,
          2.0: 0.3,
          4.0 + 1e-10: 0.2
      }, {
          1.0 + 1e-10: 0.1,
          2.0: 0.3 + 1e-10,
          4.0: 0.2 - 1e-10
      }, True),
      # Dictionary not almost equal
      ({
          1: 0.1,
          2: 0.3,
      }, {
          2: 0.3,
          4: 0.2
      }, False))
  def test_dictionary_almost_equal(self, dict1, dict2, expected_result):
    if expected_result:
      test_util.assert_dictionary_almost_equal(self, dict1, dict2)
    else:
      with self.assertRaises(AssertionError):
        test_util.assert_dictionary_almost_equal(self, dict1, dict2)

  @parameterized.parameters(
      (2, 1, True), (2, 2+1e-10, True), (2+1e-10, 2, True),
      (1, 2, False))
  def test_assert_almost_greater_equal(self, a, b, expected_result):
    if expected_result:
      test_util.assert_almost_greater_equal(self, a, b)
    else:
      with self.assertRaises(AssertionError):
        test_util.assert_almost_greater_equal(self, a, b)


if __name__ == '__main__':
  unittest.main()



# Copyright 2021 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for pld_privacy_accountant."""

import math

from absl.testing import absltest
from absl.testing import parameterized

from dp_accounting import dp_event
from dp_accounting import privacy_accountant
from dp_accounting import privacy_accountant_test
from dp_accounting.pld import pld_privacy_accountant


class PldPrivacyAccountantTest(privacy_accountant_test.PrivacyAccountantTest,
                               parameterized.TestCase):

  def _make_test_accountants(self):
    return [
        pld_privacy_accountant.PLDAccountant(
            pld_privacy_accountant.NeighborRel.ADD_OR_REMOVE_ONE),
        pld_privacy_accountant.PLDAccountant(
            pld_privacy_accountant.NeighborRel.REPLACE_ONE),
        pld_privacy_accountant.PLDAccountant(
            pld_privacy_accountant.NeighborRel.REPLACE_SPECIAL),
    ]

  @parameterized.parameters(
      (dp_event.RandomizedResponseDpEvent(0.1, 3),
       pld_privacy_accountant.NeighborRel.ADD_OR_REMOVE_ONE,
       'neighboring_relation must be `REPLACE_ONE` or `REPLACE_SPECIAL`'),
      (dp_event.LaplaceDpEvent(1.0),
       pld_privacy_accountant.NeighborRel.REPLACE_ONE,
       'neighboring_relation must be `ADD_OR_REMOVE_ONE` or `REPLACE_SPECIAL`'),
      (dp_event.PoissonSampledDpEvent(0.1, dp_event.LaplaceDpEvent(1.0)),
       pld_privacy_accountant.NeighborRel.REPLACE_ONE,
       'neighboring_relation must be `ADD_OR_REMOVE_ONE` or `REPLACE_SPECIAL`'),
      (dp_event.DiscreteLaplaceDpEvent(0.1, 1),
       pld_privacy_accountant.NeighborRel.REPLACE_ONE,
       'neighboring_relation must be `ADD_OR_REMOVE_ONE` or `REPLACE_SPECIAL`'),
      (dp_event.MixtureOfGaussiansDpEvent(1.0, [0.0, 1.0], [0.5, 0.5]),
       pld_privacy_accountant.NeighborRel.REPLACE_ONE,
       'neighboring_relation must be `ADD_OR_REMOVE_ONE` or `REPLACE_SPECIAL`'),
  )
  def test_composition_errors_for_neighboring_relation(
      self, event, neighboring_relation, error_msg):
    pld_accountant = pld_privacy_accountant.PLDAccountant(neighboring_relation)
    with self.assertRaisesRegex(privacy_accountant.UnsupportedEventError,
                                error_msg):
      pld_accountant.compose(event)

  @parameterized.named_parameters(
      ('replace_one',
       pld_privacy_accountant.NeighborRel.REPLACE_ONE, True),
      ('replace_special',
       pld_privacy_accountant.NeighborRel.REPLACE_SPECIAL, True),
      ('add_or_remove_one',
       pld_privacy_accountant.NeighborRel.ADD_OR_REMOVE_ONE, False),
  )
  def test_supports_randomized_response(
      self, neighboring_relation, supports_composition):
    pld_accountant = pld_privacy_accountant.PLDAccountant(neighboring_relation)
    event = dp_event.RandomizedResponseDpEvent(noise_parameter=0.1,
                                               num_buckets=3)
    self.assertEqual(pld_accountant.supports(event), supports_composition)

  @parameterized.named_parameters(
      ('replace_one', pld_privacy_accountant.NeighborRel.REPLACE_ONE),
      ('replace_special', pld_privacy_accountant.NeighborRel.REPLACE_SPECIAL),
  )
  def test_randomized_response_with_single_bucket(self, neighboring_relation):
    accountant = pld_privacy_accountant.PLDAccountant(neighboring_relation)
    accountant.compose(dp_event.RandomizedResponseDpEvent(0.0, 1))
    self.assertEqual(accountant.get_delta(0), 0)
    self.assertEqual(accountant.get_epsilon(0), 0)

  def test_randomized_response_with_zero_noise_parameter(self):
    accountant = pld_privacy_accountant.PLDAccountant(
        pld_privacy_accountant.NeighborRel.REPLACE_SPECIAL)
    accountant.compose(dp_event.RandomizedResponseDpEvent(0.0, 3))
    self.assertEqual(accountant.get_delta(1.0), 1)
    self.assertEqual(accountant.get_epsilon(0.01), math.inf)

  @parameterized.parameters(
      (pld_privacy_accountant.NeighborRel.REPLACE_ONE,
       4 / (3 + math.exp(1)), 4, 1.0),
      (pld_privacy_accountant.NeighborRel.REPLACE_SPECIAL,
       (4 - math.exp(1)) / 3, 4, 1.0),
      (pld_privacy_accountant.NeighborRel.REPLACE_SPECIAL,
       math.exp(-1), 2, 1.0),
  )
  def test_randomized_response(
      self, neighboring_relation,
      noise_parameter, num_buckets, expected_epsilon):
    accountant = pld_privacy_accountant.PLDAccountant(neighboring_relation)
    accountant.compose(
        dp_event.RandomizedResponseDpEvent(noise_parameter, num_buckets))
    self.assertAlmostEqual(accountant.get_delta(expected_epsilon), 0.0)

  @parameterized.product(
      event=[
          dp_event.GaussianDpEvent(1.0),
          dp_event.SelfComposedDpEvent(dp_event.GaussianDpEvent(1.0), 6),
          dp_event.ComposedDpEvent(
              [dp_event.GaussianDpEvent(1.0), dp_event.GaussianDpEvent(2.0)]
          ),
          dp_event.PoissonSampledDpEvent(0.1, dp_event.GaussianDpEvent(1.0)),
          dp_event.ComposedDpEvent([
              dp_event.PoissonSampledDpEvent(0.1,
                                             dp_event.GaussianDpEvent(1.0)),
              dp_event.GaussianDpEvent(2.0),
          ]),
      ],
      neighboring_relation=[
          pld_privacy_accountant.NeighborRel.ADD_OR_REMOVE_ONE,
          pld_privacy_accountant.NeighborRel.REPLACE_ONE,
          pld_privacy_accountant.NeighborRel.REPLACE_SPECIAL,
      ],
  )
  def test_supports_gaussian(self, event, neighboring_relation):
    pld_accountant = pld_privacy_accountant.PLDAccountant(neighboring_relation)
    self.assertTrue(pld_accountant.supports(event))

  def test_poisson_subsampling_laplace_not_supported_for_replace_one(self):
    pld_accountant = pld_privacy_accountant.PLDAccountant(
        pld_privacy_accountant.NeighborRel.REPLACE_ONE)
    event = dp_event.PoissonSampledDpEvent(0.1, dp_event.LaplaceDpEvent(1.0))
    self.assertFalse(pld_accountant.supports(event))

  @parameterized.parameters(
      (pld_privacy_accountant.NeighborRel.REPLACE_ONE, True),
      (pld_privacy_accountant.NeighborRel.REPLACE_SPECIAL, True),
      (pld_privacy_accountant.NeighborRel.ADD_OR_REMOVE_ONE, False),
  )
  def test_supports_subsampled_gaussian_and_rr_composition(
      self, neighboring_relation, supports_composition):
    pld_accountant = pld_privacy_accountant.PLDAccountant(neighboring_relation)
    event = dp_event.ComposedDpEvent([
        dp_event.PoissonSampledDpEvent(0.1, dp_event.GaussianDpEvent(1.0)),
        dp_event.RandomizedResponseDpEvent(noise_parameter=0.1, num_buckets=3)
    ])
    self.assertEqual(pld_accountant.supports(event), supports_composition)

  @parameterized.parameters(0, -1)
  def test_non_positive_composition_value_error(self, count):
    event = dp_event.GaussianDpEvent(1.0)
    accountant = pld_privacy_accountant.PLDAccountant()
    with self.assertRaises(ValueError):
      accountant.compose(event, count)

  @parameterized.parameters(
      dp_event.GaussianDpEvent(0),
      dp_event.LaplaceDpEvent(0),
      dp_event.DiscreteLaplaceDpEvent(0, 1),
      dp_event.PoissonSampledDpEvent(0.1, dp_event.GaussianDpEvent(0)),
      dp_event.PoissonSampledDpEvent(0.1, dp_event.LaplaceDpEvent(0)),
      dp_event.MixtureOfGaussiansDpEvent(0, [1, 2], [0.5, 0.5]),
  )
  def test_additive_noise_mechanisms_with_zero_noise_multiplier(self, event):
    accountant = pld_privacy_accountant.PLDAccountant()
    accountant.compose(event)
    self.assertEqual(accountant.get_delta(1.0), 1)
    self.assertEqual(accountant.get_epsilon(0.01), math.inf)

  @parameterized.parameters(
      dp_event.PoissonSampledDpEvent(0, dp_event.GaussianDpEvent(1)),
      dp_event.PoissonSampledDpEvent(0, dp_event.LaplaceDpEvent(1)),
      dp_event.PoissonSampledDpEvent(0, dp_event.GaussianDpEvent(0)),
      dp_event.PoissonSampledDpEvent(0, dp_event.LaplaceDpEvent(0)),
      dp_event.MixtureOfGaussiansDpEvent(1, [0], [1.0]),
      dp_event.MixtureOfGaussiansDpEvent(0, [0], [1.0]),
  )
  def test_poisson_subsampling_with_zero_probability(self, event):
    accountant = pld_privacy_accountant.PLDAccountant()
    accountant.compose(event)
    self.assertEqual(accountant.get_delta(0), 0)
    self.assertEqual(accountant.get_epsilon(0), 0)

  def test_gaussian_basic(self):
    gaussian_event = dp_event.GaussianDpEvent(noise_multiplier=math.sqrt(3))
    accountant = pld_privacy_accountant.PLDAccountant()
    accountant.compose(gaussian_event, 1)
    accountant.compose(gaussian_event, 2)

    exact_epsilon = 1
    exact_delta = 0.126936
    self.assertAlmostEqual(
        accountant.get_delta(exact_epsilon), exact_delta, delta=1e-3)
    self.assertAlmostEqual(
        accountant.get_epsilon(exact_delta), exact_epsilon, delta=1e-3)

  def test_poisson_subsampled_gaussian(self):
    subsampled_gaussian_event = dp_event.PoissonSampledDpEvent(
        0.2, dp_event.GaussianDpEvent(noise_multiplier=0.5))
    accountant = pld_privacy_accountant.PLDAccountant()
    accountant.compose(subsampled_gaussian_event, 1)
    accountant.compose(subsampled_gaussian_event, 2)

    exact_epsilon = 1
    expected_delta = 0.15594
    self.assertAlmostEqual(
        accountant.get_delta(exact_epsilon), expected_delta, delta=1e-3)
    self.assertAlmostEqual(
        accountant.get_epsilon(expected_delta), exact_epsilon, delta=1e-3)

  def test_self_composed_subsampled_gaussian(self):
    event = dp_event.SelfComposedDpEvent(
        dp_event.PoissonSampledDpEvent(0.2, dp_event.GaussianDpEvent(0.5)), 3)
    accountant = pld_privacy_accountant.PLDAccountant()
    accountant.compose(event)

    exact_epsilon = 1
    expected_delta = 0.15594
    self.assertAlmostEqual(
        accountant.get_delta(exact_epsilon), expected_delta, delta=1e-3)
    self.assertAlmostEqual(
        accountant.get_epsilon(expected_delta), exact_epsilon, delta=1e-3)

  def test_laplace_basic(self):
    first_laplace_event = dp_event.LaplaceDpEvent(noise_multiplier=1)
    second_laplace_event = dp_event.LaplaceDpEvent(noise_multiplier=2)
    accountant = pld_privacy_accountant.PLDAccountant()
    accountant.compose(first_laplace_event, 3)
    accountant.compose(second_laplace_event, 2)

    expected_epsilon = 4
    expected_delta = 1e-14  # expected delta is not 0 due to truncation in
    # self composition
    self.assertAlmostEqual(
        accountant.get_delta(expected_epsilon), expected_delta, delta=1e-6)
    self.assertAlmostEqual(
        accountant.get_epsilon(expected_delta), expected_epsilon, delta=1e-6)

  def test_poisson_subsampled_laplace(self):
    subsampled_laplace_event = dp_event.PoissonSampledDpEvent(
        0.2, dp_event.LaplaceDpEvent(noise_multiplier=0.5))
    accountant = pld_privacy_accountant.PLDAccountant()
    accountant.compose(subsampled_laplace_event, 1)
    accountant.compose(subsampled_laplace_event, 2)

    exact_epsilon = 2.46964
    expected_delta = 1e-14  # expected delta is not 0 due to truncation in
    # self composition
    self.assertAlmostEqual(
        accountant.get_delta(exact_epsilon), expected_delta, delta=1e-6)
    self.assertAlmostEqual(
        accountant.get_epsilon(expected_delta), exact_epsilon, delta=1e-3)

  def test_discrete_laplace_basic(self):
    first_discrete_laplace_event = dp_event.DiscreteLaplaceDpEvent(
        noise_parameter=1, sensitivity=1
    )
    second_discrete_laplace_event = dp_event.DiscreteLaplaceDpEvent(
        noise_parameter=0.5, sensitivity=1
    )
    accountant = pld_privacy_accountant.PLDAccountant()
    accountant.compose(first_discrete_laplace_event, 3)
    accountant.compose(second_discrete_laplace_event, 2)
    expected_epsilon = 4
    expected_delta = 0.0
    self.assertAlmostEqual(
        accountant.get_delta(expected_epsilon), expected_delta, delta=1e-6)
    self.assertAlmostEqual(
        accountant.get_epsilon(expected_delta), expected_epsilon, delta=1e-6)

  def test_mixture_of_gaussians_basic(self):
    first_mog_event = dp_event.MixtureOfGaussiansDpEvent(
        standard_deviation=1.0,
        sensitivities=[0.0, 1.0, 2.0],
        sampling_probs=[0.2, 0.6, 0.2],
    )
    second_mog_event = dp_event.MixtureOfGaussiansDpEvent(
        standard_deviation=2.0,
        sensitivities=[0.0, 2.0, 3.0],
        sampling_probs=[0.33, 0.34, 0.33],
    )
    accountant = pld_privacy_accountant.PLDAccountant()
    accountant.compose(first_mog_event, 3)
    accountant.compose(second_mog_event, 2)

    expected_epsilon = 17.193115
    expected_delta = 1e-6
    self.assertAlmostEqual(
        accountant.get_delta(expected_epsilon), expected_delta, delta=1e-12
    )
    self.assertAlmostEqual(
        accountant.get_epsilon(expected_delta), expected_epsilon, delta=1e-6
    )


if __name__ == '__main__':
  absltest.main()



# Copyright 2020 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Helper functions for privacy accounting across queries."""

import math
from typing import Callable, Optional

from scipy import special

from dp_accounting.pld import common
from dp_accounting.pld import privacy_loss_distribution
from dp_accounting.pld import privacy_loss_mechanism


def get_smallest_parameter(
    privacy_parameters: common.DifferentialPrivacyParameters,
    num_queries: int,
    privacy_loss_distribution_constructor: Callable[
        [float], privacy_loss_distribution.PrivacyLossDistribution],
    search_parameters: common.BinarySearchParameters
    ) -> Optional[float]:
  """Finds smallest parameter for which the mechanism satisfies desired privacy.

  This function computes the smallest "parameter" for which the corresponding
  mechanism, when run a specified number of times, satisfies a given privacy
  level. It is assumed that, when the parameter increases, the mechanism becomes
  more private.

  Args:
    privacy_parameters: The desired privacy guarantee.
    num_queries: Number of times the mechanism will be invoked.
    privacy_loss_distribution_constructor: A function that takes in a parameter
      and returns the privacy loss distribution for the corresponding mechanism
      for the given parameter.
    search_parameters: Parameters used for binary search.

  Returns:
    Smallest parameter for which the corresponding mechanism with that
    parameter, when applied the given number of times, satisfies the desired
    privacy guarantee. When no parameter in the given range satisfies this,
    return None.
  """

  def get_delta_for_parameter(parameter):
    pld_single_query = privacy_loss_distribution_constructor(parameter)
    pld_all_queries = pld_single_query.self_compose(num_queries)
    return pld_all_queries.get_delta_for_epsilon(privacy_parameters.epsilon)

  return common.inverse_monotone_function(get_delta_for_parameter,
                                          privacy_parameters.delta,
                                          search_parameters)


def get_smallest_laplace_noise(
    privacy_parameters: common.DifferentialPrivacyParameters,
    num_queries: int = 1,
    sensitivity: float = 1) -> float:
  """Finds smallest Laplace noise for which the mechanism satisfies desired privacy.

  Args:
    privacy_parameters: The desired privacy guarantee.
    num_queries: Number of times the mechanism will be invoked.
    sensitivity: The l1 sensitivity of each query.

  Returns:
    Smallest parameter for which the Laplace mechanism with this parameter, when
    applied the given number of times, satisfies the desired privacy guarantee.
  """

  def privacy_loss_distribution_constructor(parameter):
    # Setting value_discretization_interval equal to
    # 0.01 * epsilon / num_queries ensures that the resulting parameter is not
    # (epsilon', delta)-DP for epsilon' less than  0.99 * epsilon / num_queries.
    # Below, we use 10x larger discretization, since we are using the tighter
    # connect-the-dots algorithm for discretizing PLDs.
    # This is a heuristic for getting a reasonable pessimistic estimate for the
    # noise parameter.
    return privacy_loss_distribution.from_laplace_mechanism(
        parameter,
        sensitivity=sensitivity,
        value_discretization_interval=0.1 * privacy_parameters.epsilon /
        num_queries)

  # Laplace mechanism with parameter sensitivity * num_queries / epsilon is
  # epsilon-DP (for num_queries queries).
  search_parameters = common.BinarySearchParameters(
      0, num_queries * sensitivity / privacy_parameters.epsilon)

  parameter = get_smallest_parameter(privacy_parameters, num_queries,
                                     privacy_loss_distribution_constructor,
                                     search_parameters)
  if parameter is None:
    parameter = num_queries * sensitivity / privacy_parameters.epsilon
  return parameter


def get_smallest_discrete_laplace_noise(
    privacy_parameters: common.DifferentialPrivacyParameters,
    num_queries: int = 1,
    sensitivity: int = 1) -> float:
  """Finds smallest discrete Laplace noise for which the mechanism satisfies desired privacy.

  Note that from the way discrete Laplace distribution is defined, the amount of
  noise decreases as the parameter increases. (In other words, the mechanism
  becomes less private as the parameter increases.) As a result, the output will
  be the largest parameter (instead of smallest as in Laplace).

  Args:
    privacy_parameters: The desired privacy guarantee.
    num_queries: Number of times the mechanism will be invoked.
    sensitivity: The l1 sensitivity of each query.

  Returns:
    Largest parameter for which the discrete Laplace mechanism with this
    parameter, when applied the given number of times, satisfies the desired
    privacy guarantee.
  """

  # Search for inverse of the parameter instead of the parameter itself.
  def privacy_loss_distribution_constructor(inverse_parameter):
    parameter = 1 / inverse_parameter
    # Setting value_discretization_interval equal to parameter because the
    # privacy loss of discrete Laplace mechanism is always divisible by the
    # parameter.
    return privacy_loss_distribution.from_discrete_laplace_mechanism(
        parameter,
        sensitivity=sensitivity,
        value_discretization_interval=parameter)

  # discrete Laplace mechanism with parameter
  # epsilon / (sensitivity * num_queries) is epsilon-DP (for num_queries
  # queries).
  search_parameters = common.BinarySearchParameters(
      0, num_queries * sensitivity / privacy_parameters.epsilon)

  inverse_parameter = get_smallest_parameter(
      privacy_parameters, num_queries, privacy_loss_distribution_constructor,
      search_parameters)
  if inverse_parameter is None:
    parameter = privacy_parameters.epsilon / (num_queries * sensitivity)
  else:
    parameter = 1 / inverse_parameter
  return parameter


def get_smallest_gaussian_noise(
    privacy_parameters: common.DifferentialPrivacyParameters,
    num_queries: int = 1,
    sensitivity: float = 1) -> float:
  """Finds smallest noise for desired privacy of Gaussian mechanism.

  Args:
    privacy_parameters: The desired privacy guarantee.
    num_queries: Number of times the mechanism will be invoked.
    sensitivity: The l2 sensitivity of each query.

  Returns:
    Smallest standard deviation for which the Gaussian mechanism with this std,
    when applied the given number of times, satisfies the desired privacy
    guarantee.
  """
  # The l2 sensitivity grows as square root of the number of queries
  return privacy_loss_mechanism.GaussianPrivacyLoss.from_privacy_guarantee(
      privacy_parameters,
      sensitivity=sensitivity * math.sqrt(num_queries)).standard_deviation


def get_smallest_subsampled_gaussian_noise(
    privacy_parameters: common.DifferentialPrivacyParameters,
    num_queries: int = 1,
    sensitivity: int = 1,
    sampling_prob: float = 1.0) -> float:
  """Finds smallest noise for desired privacy of subsampled Gaussian mechanism.

  Args:
    privacy_parameters: The desired privacy guarantee.
    num_queries: Number of times the mechanism will be invoked.
    sensitivity: The l2 sensitivity of each query.
    sampling_prob: Subsampling probability of the mechanism.

  Returns:
    Smallest standard deviation for which the Poisson subsampled Gaussian
    mechanism with this standard deviation with specified sub-sampling
    probability , when applied the given number of times, satisfies the
    desired privacy guarantee.

  Raises:
    RuntimeError if parameter value is not found.
  """

  def privacy_loss_distribution_constructor(parameter):
    # Setting value_discretization_interval equal to the smaller of 1e-3 and
    # 0.01 * epsilon / num_queries ensures that the resulting parameter is not
    # (epsilon', delta)-DP for epsilon' less than  0.99 * epsilon / num_queries.
    # Below, we use 10x larger discretization, since we are using the tighter
    # connect-the-dots algorithm for discretizing PLDs.
    # This is a heuristic for getting a reasonable pessimistic estimate for the
    # noise parameter.
    return privacy_loss_distribution.from_gaussian_mechanism(
        parameter,
        sensitivity=sensitivity,
        value_discretization_interval=min(
            1e-3, 0.1 * privacy_parameters.epsilon / num_queries),
        sampling_prob=sampling_prob)

  upper_bound = get_smallest_gaussian_noise(
      privacy_parameters, num_queries, sensitivity)
  search_parameters = common.BinarySearchParameters(0, upper_bound)
  parameter = get_smallest_parameter(privacy_parameters,
                                     num_queries,
                                     privacy_loss_distribution_constructor,
                                     search_parameters)
  return parameter if parameter is not None else upper_bound


def advanced_composition(
    privacy_parameters: common.DifferentialPrivacyParameters,
    num_queries: int,
    total_delta: float) -> Optional[float]:
  """Computes total DP parameters after applying an algorithm with given privacy parameters multiple times.

  Using the optimal advanced composition theorem, Theorem 3.3 from the paper
  Kairouz, Oh, Viswanath. "The Composition Theorem for Differential Privacy",
  to compute the total DP parameters given that we are applying an algorithm
  with a given privacy parameters for a given number of times.

  Note that we can compute this alternatively from PrivacyLossDistribution
  by invoking from_privacy_parameters and applying the given number of
  composition. When setting value_discretization_interval appropriately, these
  two approaches should coincide but using the advanced composition theorem
  directly is less computational intensive.

  Args:
    privacy_parameters: The privacy guarantee of a single query.
    num_queries: Number of times the algorithm is invoked.
    total_delta: The target value of total delta of the privacy parameters for
      the multiple runs of the algorithm.

  Returns:
    total_epsilon such that, when applying the algorithm the given number of
    times, the result is still (total_epsilon, total_delta)-DP.

    None when the total_delta is less than 1 - (1 - delta)^num_queries, for
    which no guarantee of (total_epsilon, total_delta)-DP is possible for any
    value of total_epsilon.
  """
  epsilon = privacy_parameters.epsilon
  delta = privacy_parameters.delta
  k = num_queries

  # The calculation follows Theorem 3.3 of https://arxiv.org/pdf/1311.0776.pdf
  for i in range(k // 2, -1, -1):
    delta_i = 0
    for l in range(i):
      delta_i += special.binom(k, l) * (
          math.exp(epsilon * (k - l)) - math.exp(epsilon * (k - 2 * i + l)))
    delta_i /= ((1 + math.exp(epsilon))**k)
    if 1 - ((1 - delta) ** k) * (1 - delta_i) <= total_delta:
      return epsilon * (k - 2 * i)
  return None


def get_smallest_epsilon_from_advanced_composition(
    total_privacy_parameters: common.DifferentialPrivacyParameters,
    num_queries: int,
    delta: float = 0) -> Optional[float]:
  """Computes DP parameters that after a certain number of queries remain DP with given parameters.

  Using the optimal advanced composition theorem, Theorem 3.3 from the paper
  Kairouz, Oh, Viswanath. "The Composition Theorem for Differential Privacy",
  to compute DP parameter for an algorithm, so that when applied a given number
  of times it remains DP with given privacy parameters.

  Args:
    total_privacy_parameters: The desired privacy guarantee after applying the
      algorithm a given number of times.
    num_queries: Number of times the algorithm is invoked.
    delta: The value of DP parameter delta for the algorithm.

  Returns:
    epsilon such that if an algorithm is (epsilon, delta)-DP, then applying it
    the given number of times remains DP with total_privacy_parameters.

    None when total_privacy_parameters.delta is less than
    1 - (1 - delta)^num_queries for which no guarantee of
    total_privacy_parameters DP is possible for any value of epsilon.
  """
  if 1 - ((1 - delta) ** num_queries) > total_privacy_parameters.delta:
    return None

  search_parameters = common.BinarySearchParameters(
      total_privacy_parameters.epsilon / num_queries,
      total_privacy_parameters.epsilon)

  def get_total_epsilon_for_epsilon(epsilon):
    privacy_parameters = common.DifferentialPrivacyParameters(epsilon, delta)
    return advanced_composition(privacy_parameters, num_queries,
                                total_privacy_parameters.delta)

  return common.inverse_monotone_function(
      get_total_epsilon_for_epsilon,
      total_privacy_parameters.epsilon,
      search_parameters,
      increasing=True)



# Copyright 2020 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Common classes and functions for the accounting library."""

import dataclasses
import math
from typing import Callable, List, Mapping, Optional, Tuple, Union

import numpy as np
from scipy import fft
from scipy import signal
from scipy import special

ArrayLike = Union[np.ndarray, List[float]]


@dataclasses.dataclass
class DifferentialPrivacyParameters(object):
  """Representation of the differential privacy parameters of a mechanism.

  Attributes:
    epsilon: the epsilon in (epsilon, delta)-differential privacy.
    delta: the delta in (epsilon, delta)-differential privacy.
  """
  epsilon: float
  delta: float = 0

  def __post_init__(self):
    if self.epsilon < 0:
      raise ValueError(f'epsilon should be positive: {self.epsilon}')
    if self.delta < 0 or self.delta > 1:
      raise ValueError(f'delta should be between 0 and 1: {self.delta}')


@dataclasses.dataclass
class BinarySearchParameters(object):
  """Parameters used for binary search.

  Attributes:
    upper_bound: An upper bound on the binary search range.
    lower_bound: A lower bound on the binary search range.
    initial_guess: An initial guess to start the search with. Must be positive.
      When this guess is close to the true value, it can help make the binary
      search faster.
    tolerance: An acceptable error on the returned value.
    discrete: Whether the search is over integers.
  """
  lower_bound: float
  upper_bound: float
  initial_guess: Optional[float] = None
  tolerance: float = 1e-7
  discrete: bool = False


def inverse_monotone_function(func: Callable[[float], float],
                              value: float,
                              search_parameters: BinarySearchParameters,
                              increasing: bool = False) -> Optional[float]:
  """Inverse a monotone function.

  Args:
    func: The function to be inversed.
    value: The desired value of the function.
    search_parameters: Parameters used for binary search.
    increasing: Whether the function is monotonically increasing.

  Returns:
    x such that func(x) is no more than value, when such x exists. It is
    guaranteed that the returned x is within search_parameters.tolerance of the
    smallest (for monotonically decreasing func) or the largest (for
    monotonically increasing func) such x. When no such x exists within the
    given range, returns None.
  """
  lower_x = search_parameters.lower_bound
  upper_x = search_parameters.upper_bound
  initial_guess_x = search_parameters.initial_guess

  if increasing:
    check = lambda func_value, target_value: func_value <= target_value
    if lower_x != -math.inf and func(lower_x) > value:
      return None
  else:
    check = lambda func_value, target_value: func_value > target_value
    if upper_x != math.inf and func(upper_x) > value:
      return None

  if initial_guess_x is not None:
    while initial_guess_x < upper_x and check(func(initial_guess_x), value):
      lower_x = initial_guess_x
      initial_guess_x *= 2
    upper_x = min(upper_x, initial_guess_x)

  if search_parameters.discrete:
    tolerance = 1
  else:
    tolerance = search_parameters.tolerance

  while upper_x - lower_x > tolerance:
    if search_parameters.discrete:
      mid_x = (upper_x + lower_x) // 2
    else:
      mid_x = (upper_x + lower_x) / 2

    if check(func(mid_x), value):
      lower_x = mid_x
    else:
      upper_x = mid_x

  if increasing:
    return lower_x
  else:
    return upper_x


def dictionary_to_list(
    input_dictionary: Mapping[int, float]) -> Tuple[int, List[float]]:
  """Converts an integer-keyed dictionary into an list.

  Args:
    input_dictionary: A dictionary whose keys are integers.

  Returns:
    A tuple of an integer offset and a list result_list. The offset is the
    minimum value of the input dictionary. result_list has length equal to the
    difference between the maximum and minimum values of the input dictionary.
    result_list[i] is equal to dictionary[offset + i] and is zero if offset + i
    is not a key in the input dictionary.
  """
  offset = min(input_dictionary)
  max_val = max(input_dictionary)
  result_list = [input_dictionary.get(i, 0) for i in range(offset, max_val + 1)]
  return (offset, result_list)


def list_to_dictionary(input_list: List[float],
                       offset: int,
                       tail_mass_truncation: float = 0) -> Mapping[int, float]:
  """Converts a list into an integer-keyed dictionary, with a specified offset.

  Args:
    input_list: An input list.
    offset: The offset in the key of the output dictionary
    tail_mass_truncation: an upper bound on the tails of the input list that
      might be truncated.

  Returns:
    A dictionary whose value at key is equal to input_list[key - offset]. If
    input_list[key - offset] is less than or equal to zero, it is not included
    in the dictionary.
  """
  lower_truncation_index = 0
  lower_truncation_mass = 0
  while lower_truncation_index < len(input_list):
    lower_truncation_mass += input_list[lower_truncation_index]
    if lower_truncation_mass > tail_mass_truncation / 2:
      break
    lower_truncation_index += 1

  upper_truncation_index = len(input_list) - 1
  upper_truncation_mass = 0
  while upper_truncation_index >= 0:
    upper_truncation_mass += input_list[upper_truncation_index]
    if upper_truncation_mass > tail_mass_truncation / 2:
      break
    upper_truncation_index -= 1

  result_dictionary = {}
  for i in range(lower_truncation_index, upper_truncation_index + 1):
    if input_list[i] > 0:
      result_dictionary[i + offset] = input_list[i]
  return result_dictionary


def convolve_dictionary(dictionary1: Mapping[int, float],
                        dictionary2: Mapping[int, float],
                        tail_mass_truncation: float = 0) -> Mapping[int, float]:
  """Computes a convolution of two dictionaries.

  Args:
    dictionary1: The first dictionary whose keys are integers.
    dictionary2: The second dictionary whose keys are integers.
    tail_mass_truncation: an upper bound on the tails of the output that might
      be truncated.

  Returns:
    The dictionary where for each key its corresponding value is the sum, over
    all key1, key2 such that key1 + key2 = key, of dictionary1[key1] times
    dictionary2[key2]
  """

  # Convert the dictionaries to lists.
  min1, list1 = dictionary_to_list(dictionary1)
  min2, list2 = dictionary_to_list(dictionary2)

  # Compute the convolution of the two lists.
  result_list = signal.fftconvolve(list1, list2)

  # Convert the list back to a dictionary and return
  return list_to_dictionary(
      result_list, min1 + min2, tail_mass_truncation=tail_mass_truncation)


def compute_self_convolve_bounds(
    input_list: List[float],
    num_times: int,
    tail_mass_truncation: float = 0,
    orders: Optional[List[float]] = None) -> Tuple[int, int]:
  """Computes truncation bounds for convolution using Chernoff bound.

  Args:
    input_list: The input list to be convolved.
    num_times: The number of times the list is to be convolved with itself.
    tail_mass_truncation: an upper bound on the tails of the output that might
      be truncated.
    orders: a list of orders on which the Chernoff bound is applied.

  Returns:
    A pair of upper and lower bounds for which the mass of the result of
    convolution outside of this range is at most tail_mass_truncation.
  """
  upper_bound = (len(input_list) - 1) * num_times
  lower_bound = 0

  if tail_mass_truncation == 0:
    return lower_bound, upper_bound

  if orders is None:
    # Set orders so whose absolute values are not too large; otherwise, we may
    # run into numerical issues.
    orders = (
        np.concatenate((np.arange(-20, 0), np.arange(1, 21))) / len(input_list))

  # Compute log of the moment generating function at the specified orders.
  log_mgfs = [
      special.logsumexp(np.arange(len(input_list)) * order, b=input_list)
      for order in orders
  ]

  for order, log_mgf_value in zip(orders, log_mgfs):
    # Use Chernoff bound to update the upper/lower bound. See equation (5) in
    # the supplementary material.
    bound = (num_times * log_mgf_value +
             math.log(2 / tail_mass_truncation)) / order
    if order > 0:
      upper_bound = min(upper_bound, math.ceil(bound))
    if order < 0:
      lower_bound = max(lower_bound, math.floor(bound))

  return lower_bound, upper_bound


def self_convolve(input_list: ArrayLike,
                  num_times: int,
                  tail_mass_truncation: float = 0) -> Tuple[int, List[float]]:
  """Computes a convolution of the input list with itself num_times times.

  Args:
    input_list: The input list to be convolved.
    num_times: The number of times the list is to be convolved with itself.
    tail_mass_truncation: an upper bound on the tails of the output that might
      be truncated.

  Returns:
    A pair of truncation_lower_bound, output_list, where the i-th entry of
    output_list is approximately the sum, over all i_1, i_2, ..., i_num_times
    such that i_1 + i_2 + ... + i_num_times = i + truncation_lower_bound,
    of input_list[i_1] * input_list[i_2] * ... * input_list[i_num_times].
  """
  truncation_lower_bound, truncation_upper_bound = compute_self_convolve_bounds(
      input_list, num_times, tail_mass_truncation)

  # Use FFT to compute the convolution
  output_len = truncation_upper_bound - truncation_lower_bound + 1
  fast_len = fft.next_fast_len(max(output_len, len(input_list)))
  truncated_convolution_output = np.real(
      fft.ifft(fft.fft(input_list, fast_len)**num_times))

  # Discrete Fourier Transform wraps around modulo fast_len. Extract the output
  # values in the range of interest.
  output_list = np.roll(
      truncated_convolution_output, -truncation_lower_bound
  )[:output_len]

  return truncation_lower_bound, output_list


def self_convolve_dictionary(
    input_dictionary: Mapping[int, float],
    num_times: int,
    tail_mass_truncation: float = 0) -> Mapping[int, float]:
  """Computes a convolution of the input dictionary with itself num_times times.

  Args:
    input_dictionary: The input dictionary whose keys are integers.
    num_times: The number of times the dictionary is to be convolved with
      itself.
    tail_mass_truncation: an upper bound on the tails of the output that might
      be truncated.

  Returns:
    The dictionary where for each key its corresponding value is the sum, over
    all key1, key2, ..., key_num_times such that key1 + key2 + ... +
    key_num_times = key, of input_dictionary[key1] * input_dictionary[key2] *
    ... * input_dictionary[key_num_times]
  """
  min_val, input_list = dictionary_to_list(input_dictionary)
  min_val_convolution, output_list = self_convolve(
      input_list, num_times, tail_mass_truncation=tail_mass_truncation)
  return list_to_dictionary(output_list,
                            min_val * num_times + min_val_convolution)


def _log_add(a: float, b: float) -> float:
  """Returns log(exp(a) + exp(b))."""
  mn, mx = min(a, b), max(a, b)
  return mx + np.log1p(np.exp(mn - mx))


def _log_sub(a: float, b: float) -> float:
  """Returns log(exp(a) - exp(b))."""
  if b >= a:
    raise ValueError(f'a must be greater than b. Got a={a} and b={b}.')
  return a + np.log1p(-np.exp(b - a))


def log_a_times_exp_b_plus_c(a: float, b: float, c: float) -> float:
  """Computes log(a * exp(b) + c)."""
  if a == 0:
    return np.log(c)
  if a < 0:
    if c <= 0:
      raise ValueError(f'a exp(b) + c must be positive: {a}, {b}, {c}.')
    return _log_sub(np.log(c), np.log(-a) + b)
  if b == 0:
    return np.log(a + c)
  d = b + np.log(a)
  if c == 0:
    return d
  elif c < 0:
    return _log_sub(d, np.log(-c))
  else:
    return _log_add(d, np.log(c))


def log_sinh(a: float) -> float:
  """Returns log(sinh(a)) = log((exp(a) - exp(-a)) / 2)."""
  if a < 0:
    raise ValueError(f'a must be non-negative. Got a={a}.')
  if a == 0:
    return 0
  # log(exp(a) * (1 - exp(-2a)) / 2) = a + log(1 - exp(-2a)) - log(2)
  return a + math.log1p(-math.exp(-2 * a)) - math.log(2)


def asinh_exp(a: float, b: float) -> float:
  """Returns asinh(b * exp(a)).

  Positive solution of the following quadratic equation in X = e^x:
    X^2 - 2 * b * e^a * X - 1 = 0
  Given as: X = sqrt{(b * e^a)^2 + 1} + b * e^a
              = 1 / (sqrt{(b * e^a)^2 + 1} - b * e^a)
  Thus, we get x = log(sqrt{(b * e^a)^2 + 1} + b * e^a)   [useful when b > 0]
                 = - log(sqrt{(b * e^a)^2 + 1} - b * e^a) [useful when b < 0]
                 = sign(b) * log(sqrt{(b * e^a)^2 + 1} + |b| * e^a)
  Args:
    a: The first argument.
    b: The second argument.
  """
  if b == 0.0:
    return 0.0
  sign_b, abs_b = (1, b) if b >= 0 else (-1, -b)
  log_abs_b_times_exp_a = math.log(abs_b) + a
  return sign_b * np.logaddexp(
      0.5 * np.logaddexp(0, 2 * log_abs_b_times_exp_a),
      log_abs_b_times_exp_a,
  )



# Copyright 2020 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import math
import unittest
from unittest import mock

from absl.testing import parameterized

from dp_accounting.pld import common
from dp_accounting.pld import test_util


class DifferentialPrivacyParametersTest(parameterized.TestCase):

  @parameterized.parameters((-0.1, 0.1), (1, -0.1), (1, 1.1))
  def test_epsilon_delta_value_errors(self, epsilon, delta):
    with self.assertRaises(ValueError):
      common.DifferentialPrivacyParameters(epsilon, delta)


class CommonTest(parameterized.TestCase):

  @parameterized.named_parameters(
      {
          'testcase_name': 'no_initial_guess',
          'func': (lambda x: -x),
          'value': -4.5,
          'lower_x': 0,
          'upper_x': 10,
          'initial_guess_x': None,
          'expected_x': 4.5,
          'increasing': False,
      }, {
          'testcase_name': 'with_initial_guess',
          'func': (lambda x: -x),
          'value': -5,
          'lower_x': 0,
          'upper_x': 10,
          'initial_guess_x': 2,
          'expected_x': 5,
          'increasing': False,
      }, {
          'testcase_name': 'out_of_range',
          'func': (lambda x: -x),
          'value': -5,
          'lower_x': 0,
          'upper_x': 4,
          'initial_guess_x': None,
          'expected_x': None,
          'increasing': False,
      }, {
          'testcase_name': 'infinite_upper_bound',
          'func': (lambda x: -1 / (1 / x)),
          'value': -5,
          'lower_x': 0,
          'upper_x': math.inf,
          'initial_guess_x': 2,
          'expected_x': 5,
          'increasing': False,
      }, {
          'testcase_name': 'increasing_no_initial_guess',
          'func': (lambda x: x**2),
          'value': 25,
          'lower_x': 0,
          'upper_x': 10,
          'initial_guess_x': None,
          'expected_x': 5,
          'increasing': True,
      }, {
          'testcase_name': 'increasing_with_initial_guess',
          'func': (lambda x: x**2),
          'value': 25,
          'lower_x': 0,
          'upper_x': 10,
          'initial_guess_x': 2,
          'expected_x': 5,
          'increasing': True,
      }, {
          'testcase_name': 'increasing_out_of_range',
          'func': (lambda x: x**2),
          'value': 5,
          'lower_x': 6,
          'upper_x': 10,
          'initial_guess_x': None,
          'expected_x': None,
          'increasing': True,
      }, {
          'testcase_name': 'discrete',
          'func': (lambda x: -x),
          'value': -4.5,
          'lower_x': 0,
          'upper_x': 10,
          'initial_guess_x': None,
          'expected_x': 5,
          'increasing': False,
          'discrete': True,
      })
  def test_inverse_monotone_function(self,
                                     func,
                                     value,
                                     lower_x,
                                     upper_x,
                                     initial_guess_x,
                                     expected_x,
                                     increasing,
                                     discrete=False):
    search_parameters = common.BinarySearchParameters(
        lower_x, upper_x, initial_guess=initial_guess_x, discrete=discrete)
    x = common.inverse_monotone_function(
        func, value, search_parameters, increasing=increasing)
    if expected_x is None:
      self.assertIsNone(x)
    else:
      self.assertAlmostEqual(expected_x, x)


class DictListConversionTest(parameterized.TestCase):

  @parameterized.named_parameters(
      {
          'testcase_name': 'truncate_both_sides',
          'input_list': [0.2, 0.5, 0.3],
          'offset': 1,
          'tail_mass_truncation': 0.6,
          'expected_result': {
              2: 0.5
          },
      }, {
          'testcase_name': 'truncate_lower_only',
          'input_list': [0.2, 0.5, 0.3],
          'offset': 1,
          'tail_mass_truncation': 0.4,
          'expected_result': {
              2: 0.5,
              3: 0.3
          },
      }, {
          'testcase_name': 'truncate_upper_only',
          'input_list': [0.4, 0.5, 0.1],
          'offset': 1,
          'tail_mass_truncation': 0.3,
          'expected_result': {
              1: 0.4,
              2: 0.5
          },
      }, {
          'testcase_name': 'truncate_all',
          'input_list': [0.4, 0.5, 0.1],
          'offset': 1,
          'tail_mass_truncation': 3,
          'expected_result': {},
      })
  def test_list_to_dict_truncation(self, input_list, offset,
                                   tail_mass_truncation, expected_result):
    result = common.list_to_dictionary(
        input_list, offset, tail_mass_truncation=tail_mass_truncation)
    test_util.assert_dictionary_almost_equal(self, expected_result, result)


class ConvolveTest(parameterized.TestCase):

  def test_convolve_dictionary(self):
    dictionary1 = {1: 2, 3: 4}
    dictionary2 = {2: 3, 4: 6}
    expected_result = {3: 6, 5: 24, 7: 24}
    result = common.convolve_dictionary(dictionary1, dictionary2)
    test_util.assert_dictionary_almost_equal(self, expected_result, result)

  def test_convolve_dictionary_with_truncation(self):
    dictionary1 = {1: 0.4, 2: 0.6}
    dictionary2 = {1: 0.7, 3: 0.3}
    expected_result = {3: 0.42, 4: 0.12}
    result = common.convolve_dictionary(dictionary1, dictionary2, 0.57)
    test_util.assert_dictionary_almost_equal(self, expected_result, result)

  def test_self_convolve_dictionary(self):
    inp_dictionary = {1: 2, 3: 5, 4: 6}
    expected_result = {
        3: 8,
        5: 60,
        6: 72,
        7: 150,
        8: 360,
        9: 341,
        10: 450,
        11: 540,
        12: 216
    }
    result = common.self_convolve_dictionary(inp_dictionary, 3)
    test_util.assert_dictionary_almost_equal(self, expected_result, result)

  @parameterized.parameters(([3, 5, 7], 2, [9, 30, 67, 70, 49]),
                            ([1, 3, 4], 3, [1, 9, 39, 99, 156, 144, 64]))
  def test_self_convolve_basic(self, input_list, num_times, expected_result):
    min_val, result_list = common.self_convolve(input_list, num_times)
    self.assertEqual(0, min_val)
    self.assertSequenceAlmostEqual(expected_result, result_list)

  @parameterized.parameters(([0.1, 0.4, 0.5], 3, [-1], 0.5, 2, 6),
                            ([0.2, 0.6, 0.2], 3, [1], 0.7, 0, 5))
  def test_compute_self_convolve_bounds(self, input_list, num_times, orders,
                                        tail_mass_truncation,
                                        expected_lower_bound,
                                        expected_upper_bound):
    lower_bound, upper_bound = common.compute_self_convolve_bounds(
        input_list, num_times, tail_mass_truncation, orders=orders)
    self.assertEqual(expected_lower_bound, lower_bound)
    self.assertEqual(expected_upper_bound, upper_bound)

  @parameterized.parameters(
      ([0.1, 0.4, 0.5], 3, 0.5, 2, [0.063, 0.184, 0.315, 0.301, 0.137]),
      ([0.2, 0.6, 0.2], 3, 0.7, 1, [0.08, 0.24, 0.36, 0.24, 0.08]))
  def test_compute_self_convolve_with_truncation(self, input_list, num_times,
                                                 tail_mass_truncation,
                                                 expected_min_val,
                                                 expected_result_list):
    min_val, result_list = common.self_convolve(
        input_list, num_times, tail_mass_truncation=tail_mass_truncation)
    self.assertEqual(min_val, expected_min_val)
    self.assertSequenceAlmostEqual(expected_result_list, result_list)

  @mock.patch.object(
      common, 'compute_self_convolve_bounds', return_value=(6, 6)
  )
  def test_compute_self_convolve_with_too_small_truncation(self, _):
    # When the truncation bounds returned from compute_self_convolve_bounds are
    # too small, the input should not be truncated.
    min_val, result_list = common.self_convolve([0, 0, 1], 3)
    self.assertEqual(min_val, 6)
    self.assertSequenceAlmostEqual([1], result_list)


class HelperMethodsTest(parameterized.TestCase):

  @parameterized.parameters(
      (5, 7, 3, 8.60998489),
      (0.5, 3, 0.1, 2.31676098),
      (10, 100000, 5, 100002.30258509))
  def test_log_a_times_exp_b_plus_c(self, a, b, c, expected_result):
    self.assertAlmostEqual(
        common.log_a_times_exp_b_plus_c(a, b, c), expected_result)

  def test_log_sinh_errors(self):
    with self.assertRaises(ValueError):
      common.log_sinh(-1)

  @parameterized.parameters(
      (0.0, 0.0),
      (1, math.log(math.sinh(1))),
      (100000, 100000 - math.log(2)),
  )
  def test_log_sinh(self, a, log_sinh_a):
    self.assertAlmostEqual(common.log_sinh(a), log_sinh_a)

  @parameterized.parameters(
      dict(a=1.0, b=0.0, asinh_exp_b_times_exp_a=0.0),
      dict(a=math.log(math.sinh(1)), b=1.0, asinh_exp_b_times_exp_a=1.0),
      dict(a=100000, b=0.5, asinh_exp_b_times_exp_a=100000),
      dict(a=math.log(math.sinh(1)), b=-1.0, asinh_exp_b_times_exp_a=-1.0),
      dict(a=100000, b=-0.5, asinh_exp_b_times_exp_a=-100000),
  )
  def test_asinh_exp(self, a, b, asinh_exp_b_times_exp_a):
    self.assertAlmostEqual(common.asinh_exp(a, b), asinh_exp_b_times_exp_a)

if __name__ == '__main__':
  unittest.main()



# Copyright 2021, The TensorFlow Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Builder class for ComposedDpEvent."""

from dp_accounting import dp_event


class DpEventBuilder(object):
  """Constructs a `DpEvent` representing the composition of a series of events.

  Two common use cases of the `DpEventBuilder` are 1) for producing and tracking
  a ledger of `DpEvent`s during sequential accounting using a
  `PrivacyAccountant`, and 2) for building up a description of a composite
  mechanism for subsequent batch accounting.
  """

  def __init__(self):
    # A list of (event, count) pairs.
    self._event_counts = []
    self._composed_event = None

  def compose(self, event: dp_event.DpEvent, count: int = 1):
    """Composes new event into event represented by builder.

    Args:
      event: The new event to compose.
      count: The number of times to compose the event.
    """
    if not isinstance(event, dp_event.DpEvent):
      raise TypeError('`event` must be a subclass of `DpEvent`. '
                      f'Found {type(event)}.')
    if not isinstance(count, int):
      raise TypeError(f'`count` must be an integer. Found {type(count)}.')
    if count < 1:
      raise ValueError(f'`count` must be positive. Found {count}.')

    if isinstance(event, dp_event.NoOpDpEvent):
      return
    elif isinstance(event, dp_event.SelfComposedDpEvent):
      self.compose(event.event, count * event.count)
    else:
      if self._event_counts and self._event_counts[-1][0] == event:
        new_event_count = (event, self._event_counts[-1][1] + count)
        self._event_counts[-1] = new_event_count
      else:
        self._event_counts.append((event, count))
      self._composed_event = None

  def build(self) -> dp_event.DpEvent:
    """Builds and returns the composed DpEvent represented by the builder."""
    if not self._composed_event:
      events = []
      for event, count in self._event_counts:
        if count == 1:
          events.append(event)
        else:
          events.append(dp_event.SelfComposedDpEvent(event, count))
      if not events:
        self._composed_event = dp_event.NoOpDpEvent()
      elif len(events) == 1:
        self._composed_event = events[0]
      else:
        self._composed_event = dp_event.ComposedDpEvent(events)

    return self._composed_event
